open Lex

let print_color = function
  Lex.Red		-> print_string "r"
| Lex.Green		-> print_string "g"
| Lex.Blue		-> print_string "b"
| Lex.Cyan		-> print_string "c"
| Lex.Magenta		-> print_string "m"
| Lex.Yellow		-> print_string "y"
| Lex.Black		-> print_string "k"
| Lex.White		-> print_string "w"
| Lex.UnknownColor	-> ()

let print_token = function
  Lex.Open_B		-> print_string "<B>"
| Lex.Close_B		-> print_string "</B>"
| Lex.Open_EM		-> print_string "<EM>"
| Lex.Close_EM		-> print_string "</EM>"
| Lex.Open_I		-> print_string "<I>"
| Lex.Close_I		-> print_string "</I>"
| Lex.Open_PL		-> print_string "<PL>"
| Lex.Close_PL		-> print_string "</PL>"
| Lex.Open_S		-> print_string "<S>"
| Lex.Close_S		-> print_string "</S>"
| Lex.Open_TT		-> print_string "<TT>"
| Lex.Close_TT		-> print_string "</TT>"
| Lex.Open_U		-> print_string "<U>"
| Lex.Close_U		-> print_string "</U>"
| Lex.Open_Size i	-> print_string "<"; print_int i; print_string ">"
| Lex.Close_Size i	-> print_string "</"; print_int i; print_string ">"
| Lex.Open_Color c	-> print_string "<"; print_color c; print_string ">"
| Lex.Close_Color c	-> print_string "</"; print_color c; print_string ">"
| Lex.Text s		-> print_string s
| Lex.Space s		-> print_string s
| Lex.EOF		-> print_string "{EOF}\n"

let length_token = function
  Lex.Open_B		-> 3
| Lex.Close_B		-> 4
| Lex.Open_EM		-> 4
| Lex.Close_EM		-> 5
| Lex.Open_I		-> 3
| Lex.Close_I		-> 4
| Lex.Open_PL		-> 4
| Lex.Close_PL		-> 5
| Lex.Open_S		-> 3
| Lex.Close_S		-> 4
| Lex.Open_TT		-> 4
| Lex.Close_TT		-> 5
| Lex.Open_U		-> 3
| Lex.Close_U		-> 4
| Lex.Open_Size i	-> 3
| Lex.Close_Size i	-> 4
| Lex.Open_Color c	-> 3
| Lex.Close_Color c	-> 4
| Lex.Text s		-> String.length s
| Lex.Space s		-> String.length s
| Lex.EOF		-> 0

exception NotOpenToken
let close_token = function
  Lex.Open_B		-> Lex.Close_B
| Lex.Open_EM		-> Lex.Close_EM
| Lex.Open_I		-> Lex.Close_I
| Lex.Open_PL		-> Lex.Close_PL
| Lex.Open_S		-> Lex.Close_S
| Lex.Open_TT		-> Lex.Close_TT
| Lex.Open_U		-> Lex.Close_U
| Lex.Open_Size i	-> Lex.Close_Size i
| Lex.Open_Color c	-> Lex.Close_Color c
| _ -> raise NotOpenToken


type decoration = {
  b: bool;
  em: bool;
  i: bool;
  s: bool;
  tt: bool;
  u: int;
  size: int;
  color: Lex.color
}
let print_bool = function
  true -> print_string "true"
| false -> print_string "false"
let print_decoration d =
  print_string "{";
  print_string "b="; print_bool d.b; print_string ";";
  print_string "em="; print_bool d.em; print_string ";";
  print_string "i="; print_bool d.i; print_string ";";
  print_string "s="; print_bool d.s; print_string ";";
  print_string "tt="; print_bool d.tt; print_string ";";
  print_string "u="; print_int d.u; print_string ";";
  print_string "size="; print_int d.size; print_string ";";
  print_string "color="; print_color d.color; print_string ";";
  print_string "}\n"

let space_decoration d =
  {d with s = false; em = false; i = false; b = false; color = if d.u == 0 then Lex.White else d.color}



let trans_b deco = { deco with b = true }
let trans_em deco =
  if deco.s then
    deco
  else
    { deco with em = not deco.em }
let trans_i deco = { deco with i = true }
let trans_pl deco = { deco with s = false; em = false; i = false; b = false; tt = false; u = 0 }
let trans_s deco = { deco with s = true; em = false }
let trans_tt deco = { deco with tt = true }
let trans_u deco =
  if deco.u < 3 then
    { deco with u = deco.u + 1 }
  else
    deco
let trans_size n deco = { deco with size = n }
let trans_color c deco = { deco with color = c }

let init_deco =  {
  b = false;
  em = false;
  i = false;
  s = false;
  tt = false;
  u = 0;
  size = -1;
  color = Lex.UnknownColor
}

type state = {
  deco: decoration;
  deco_tok_stack: (decoration * Lex.token) list;
  after_space: decoration option;
  output_len: int;
  output_buf: Lex.token Que.t;
}
let print_list p = function
  [] -> print_string "[]"
| l ->
  let sep = ref "[" in
  List.iter (fun d -> print_string !sep; sep := ";"; p d) l;
  print_string "]\n"
let print_option p = function
  None -> print_string "None"
| Some d -> print_string "Some "; p d

let print_state s =
  print_string "{";
  print_string "deco="; print_decoration s.deco; print_string ";";
  print_string "deco_tok_stack="; print_list (fun (d,t) -> print_decoration d; print_string ","; print_token t) s.deco_tok_stack; print_string ";";
  print_string "after_space="; print_option print_decoration s.after_space; print_string ";";
  print_string "output_len="; print_int s.output_len; print_string ";";
  print_string "output_buf="; print_list print_token s.output_buf; print_string ";";
  print_string "}\n"

let init_state = {
  deco = init_deco;
  deco_tok_stack = [];
  after_space = None;
  output_len = 0;
  output_buf = Que.empty;
}

let top_deco { deco_tok_stack = (deco,_)::_} = deco

let push_deco trans closetok state =
  {state with
    deco = trans state.deco;
    deco_tok_stack = (state.deco, closetok)::state.deco_tok_stack;
  }

exception CannotPop
let pop_deco = function
  ({deco_tok_stack = (top,_)::rest} as state) -> {state with deco = top; deco_tok_stack = rest}
| _ -> raise CannotPop

let enqueue_output tok state =
  {state with
    output_buf = Que.enqueue tok state.output_buf;
    output_len = state.output_len + length_token tok}

let state_trans state tok =
let state = enqueue_output tok state in
let modify_after_space state2 =
  if (space_decoration state.deco) <> (space_decoration state2.deco) then
    {state2 with after_space = None}
  else
    state2
in
match tok with
  Open_B	-> modify_after_space (push_deco trans_b Close_B state)
| Close_B	-> modify_after_space (pop_deco state)
| Open_EM	-> modify_after_space (push_deco trans_em Close_EM state)
| Close_EM	-> modify_after_space (pop_deco state)
| Open_I	-> modify_after_space (push_deco trans_i Close_I state)
| Close_I	-> modify_after_space (pop_deco state)
| Open_PL	-> modify_after_space (push_deco trans_pl Close_PL state)
| Close_PL	-> modify_after_space (pop_deco state)
| Open_S	-> modify_after_space (push_deco trans_s Close_S state)
| Close_S	-> modify_after_space (pop_deco state)
| Open_TT	-> modify_after_space (push_deco trans_tt Close_TT state)
| Close_TT	-> modify_after_space (pop_deco state)
| Open_U	-> modify_after_space (push_deco trans_u Close_U state)
| Close_U	-> modify_after_space (pop_deco state)
| Open_Size n	-> modify_after_space (push_deco (trans_size n) (Close_Size n) state)
| Close_Size n	-> modify_after_space (pop_deco state)
| Open_Color c	-> modify_after_space (push_deco (trans_color c) (Close_Color c) state)
| Close_Color c	-> modify_after_space (pop_deco state)
| Text s	-> {state with after_space = None}
| Space s	-> state
| EOF		-> state

let rec state_finish state =
match state with
  {deco_tok_stack = []} -> state
| {deco_tok_stack = (d,tok)::rest} -> state_finish (enqueue_output tok {state with deco_tok_stack = rest})


let state_can_accept state tok =
match state with
  {deco_tok_stack = (_,closetoken)::_} when tok == closetoken -> true
| _ ->
  (match tok with
    Open_B	-> true
  | Close_B	-> false
  | Open_EM	-> true
  | Close_EM	-> false
  | Open_I	-> true
  | Close_I	-> false
  | Open_PL	-> true
  | Close_PL	-> false
  | Open_S	-> true
  | Close_S	-> false
  | Open_TT	-> true
  | Close_TT	-> false
  | Open_U	-> true
  | Close_U	-> false
  | Open_Size n	-> true
  | Close_Size n	-> false
  | Open_Color c	-> true
  | Close_Color c	-> false
  | Text s	-> true
  | Space s	-> true
  | EOF		-> true)

let state_output deco state tok =
match tok with
  Text _ -> state_trans state tok 
| Space s ->
  (let deco = space_decoration deco in
  if deco.tt then
    state_trans state tok
  else
  match state.after_space with
    None ->
    (try
      let i = String.index s ' ' in
      {(state_trans state (Text (String.sub s 0 (i+1)))) with after_space = Some deco} 
    with Not_found ->
      {(state_trans state tok) with after_space = Some deco})
  | Some deco2 -> state)
| _ -> state_trans state tok

let branch_state asis_state tok state =
  (if state_can_accept state tok then
    (match tok with
      Text s ->
        if asis_state.deco = state.deco then
	  [state_output state.deco state tok]
	else
	  []
    | Space s ->
        if (space_decoration asis_state.deco) = (space_decoration state.deco) then
	  [state_output asis_state.deco state tok]
	else
	  []
    | _ -> [state_trans state tok])
  else
    []) @
  (match tok with
  | Text _	->
    let deco = asis_state.deco in
    let _, tokss = (List.fold_left
      (fun (toks, tokss) (deco2,tok2) ->
	(tok2::toks),
	(if deco = deco2 then
	  (List.rev (tok2::toks))::tokss
	else
	  tokss))
      ([],[])
      state.deco_tok_stack) in
    (List.map
      (fun s -> (state_output deco s tok))
      (List.map
	(fun toks -> List.fold_left state_trans state toks)
	tokss))
  | Space _	->
    (let deco = space_decoration asis_state.deco in
    let _, tokss = (List.fold_left
      (fun (toks, tokss) (deco2,tok2) ->
	(tok2::toks),
	(if deco = space_decoration deco2 then
	  (List.rev (tok::tok2::toks))::tokss
	else
	  tokss))
      ([],[])
      state.deco_tok_stack) in
    (List.map
      (fun s -> (state_output deco s tok))
      (List.map
	(fun toks -> List.fold_left state_trans state toks)
	tokss)))
  | _ -> [state])

let branch_states asis_state states tok = List.concat (List.map (branch_state asis_state tok) states)

let rec common_output states q1 =
match states with
  [] -> q1
| {output_buf=q2}::ss ->
  let q = Que.common_prefix q1 q2 in
  if Que.is_empty q then
    q
  else
    common_output ss q
  
let max_states = 8

let (<=>) a b = if a < b then -1 else if a == b then 0 else 1

let rec nfirst n l =
if n <= 0 then
  []
else
  match l with
    [] -> []
  | d::ds -> d::(nfirst (n-1) ds)

let main () =
  let debug = false in
  let lexbuf = Lexing.from_channel stdin in
  let tokcell = ref (Lex.token lexbuf) in
  let asis_state = ref init_state in
  let states = ref [init_state] in
  while !tokcell <> Lex.EOF
  do

    (if debug then
      (print_string "asis_state:\n";
      print_state !asis_state;
      print_string "states:\n";
      print_list print_state !states;
      print_string "token:\n";
      print_token !tokcell;
      print_string "\n\n";
      flush stdout));

    let tok = !tokcell in
    states := !asis_state::!states;
    asis_state := state_output (!asis_state).deco !asis_state tok;
    states := branch_states !asis_state !states tok;

    (let stack_length_max = (List.length (!asis_state).deco_tok_stack) + 8 in
      states := List.filter (fun {deco_tok_stack=dts} -> (List.length dts) < stack_length_max) !states);

    let hash = Hashtbl.create (2 * List.length !states) in
    List.iter (fun s1 ->
      let k1 = (s1.deco, s1.deco_tok_stack) in
      try
        let s2 = Hashtbl.find hash k1 in
	if s1.output_len < s2.output_len then
	  (Hashtbl.remove hash k1;
	  Hashtbl.add hash k1 s1)
      with Not_found ->
        Hashtbl.add hash k1 s1) !states;
    states := [];
    Hashtbl.iter (fun (d,dts) s ->
      (*
      print_decoration d;
      print_list (fun (d,t) -> print_decoration d; print_string ","; print_token t) dts; print_string ";\n";
      *)
      states := s::!states) hash;

    (*
    print_string "states:\n";
    print_list print_state !states;
    *)

    let q = common_output !states !asis_state.output_buf in
    (if not (Que.is_empty q) then
      (
      (if debug then print_string "output:\n");
      List.iter print_token q;
      let bytes = List.fold_left (fun l t -> l + length_token t) 0 q in
      let numtoks = Que.length q in
      states := List.map (fun s -> {s with output_len = s.output_len - bytes;
                                           output_buf = Que.ndequeue numtoks s.output_buf}) !states;
      asis_state := {!asis_state with output_len = (!asis_state).output_len - bytes;
                                      output_buf = Que.ndequeue numtoks (!asis_state).output_buf}));

    states := nfirst max_states (List.sort (fun a b -> a.output_len <=> b.output_len) !states);

    tokcell := Lex.token lexbuf;
  done;
  states := List.map state_finish !states;
  (if debug then
    (print_string "asis_state:\n";
    print_state !asis_state;
    print_string "states:\n";
    print_list print_state !states;
    print_string "token:\n";
    print_token !tokcell;
    print_string "\n\n";
    flush stdout));
  let state = List.hd (List.sort (fun a b -> a.output_len <=> b.output_len) !states) in
    ((if debug then print_string "output:\n");
    List.iter print_token state.output_buf)
